"""
@author: Thilogen
"""
#-------------------------------------------------------------------

import numpy as np
import math
import matplotlib.pyplot as plt

#-------------------------------------------------------------------

# Define some functions needed to calculate the necessary equations

def skew2(a):
    return a*np.array([[0, -1],[1, 0]])
def exp2(theta):
    #theta = theta.item()
    return np.array([[np.cos(theta), -np.sin(theta)],
                      [np.sin(theta), np.cos(theta)]]) 
def Ese2(theta):
    a = np.sinc(theta/np.pi)
    b = (theta/2)*(np.square(np.sinc(theta/(2*np.pi))))
    return np.array([[a, -b], [b, a]])
    
#-------------------------------------------------------------------

# Define the function that creates the real trajectory

def create_dynamics(n_time, h, om, v):
    X = np.zeros((3,n_time))
    for i in range(1,n_time):
        theta = X[0,i-1]; x = X[1:3,i-1]
        X[:,i] = np.array([theta + om * h, 
                           x[0] + np.cos(theta)* v * h, 
                           x[1] + np.sin(theta)* v * h])
    return X

#-------------------------------------------------------------------

#Define the propagate function for both filter 

def propagate(h, Xh, P, v, om, Q, config):
    theta_h = Xh[0]; x_h = Xh[1:3]; R_h = exp2(theta_h)
   
    if config == "ekf":
        A = np.zeros((3,3))
        A[1,0] = -np.sin(theta_h)*v; A[2,0] = np.cos(theta_h)*v
        
        P_p = P + h*(A @ P + P @ A.T + Q)
        
  
        Xh_p = np.array([theta_h + om * h, 
                         x_h[0] + np.cos(theta_h)* v * h, 
                         x_h[1] + np.sin(theta_h)* v * h])
   
        
    elif config == "liekf":
        A = np.zeros((3,3))
        A[2,0] = -v; A[1:3,1:3] = skew2(om)
        A = -A
        
        P_p = P + h*(A @ P + P @ A.T + Q)
        
       
        
        Xhi = np.block([[R_h, x_h.reshape(2,1)], 
                        [0,0,1]])
        
        v_lie = np.zeros((3,3))
        v_lie[0:2, 0:2] = skew2(om); v_lie[0,2] = v
        Xhi_p = Xhi + h * (Xhi @ v_lie)
        Xh_p = Xhi_p
     

    return Xh_p, P_p

#Define the update function for both filter 

def update(Xh_p, P_p, Y, N, config, step):
  
    if config == "ekf":
        Rh_p = exp2(Xh_p[0])
    elif config == "liekf":
        Rh_p = Xh_p[0:2,0:2]
    

    N_hat = Rh_p @ N @ Rh_p.T
    H = np.hstack((np.zeros((2,1)), np.eye(2)))
    
    S = H @ P_p @ H.T + N_hat
    L_n = P_p @ H.T @ np.linalg.inv(S)
    
    if config == "ekf":
        Yn = Y - Xh_p[1:3]
        Xh_u = Xh_p + (L_n @ Yn)
        
    elif config == "liekf":        
        Xhi_p = Xh_p
        
        Yn = np.block([Y, 1])
        
        inno = (np.linalg.inv(Xhi_p) @ Yn) - np.array([0,0,1])
        Ln_tilde = np.hstack((L_n, np.zeros((3,1))))
        delta = Ln_tilde @ inno
    
        E_delta = Ese2(delta[0]); R_delta = exp2(delta[0])
        exp_delta = np.block([[np.hstack((R_delta,np.array([E_delta @ delta[1:3]]).T))],
                              [np.array([0,0,1])]])
        Xhi_u = Xhi_p @ exp_delta

        # Using Atan(y,x) to find theta, where y = sin(theta) and x = cos(theta)
        thetah_u = math.atan2(Xhi_u[1,0], Xhi_u[0,0]); xh_u = Xhi_u[0:2,2]
        if (thetah_u < 0 and k > n_time/2):
            thetah_u += 2*np.pi
        Xh_u = np.hstack((thetah_u, xh_u))
    
    P_u = (np.eye(3) - L_n @ H) @ P_p

    return Xh_u, P_u

#-------------------------------------------------------------------

#Define the system initialization parameters for the simple car model

v = 1; t_circ = 40; om = 2*np.pi/t_circ; h = 0.1
n_rounds = 4/5
n_time = int(n_rounds * t_circ/h)


X = create_dynamics(n_time, h, om, v)
#-------------------------------------------------------------------

# Define the noise intialization parameters

N = np.eye(2)
Q = np.diag(((np.pi/180)**2,1e-4,1e-4))

#-------------------------------------------------------------------

filters = ["liekf","ekf"]

Xu_ekf = np.zeros((3, n_time))
Pu_ekf = np.zeros((3, 3, n_time))
Xu_liekf = np.zeros((3, n_time))
Pu_liekf = np.zeros((3, 3, n_time))

# Define heading error (degrees)

heading_error = 45

if heading_error == 45:
    Pu_ekf[:,:,0] = np.diag([(15*np.pi/180)**2, 0, 0])
    Pu_liekf[:,:,0] = np.diag([(15*np.pi/180)**2, 0, 0])
elif heading_error == 1:
    Pu_ekf[:,:,0] = np.diag([(np.pi/180)**2, 0, 0])
    Pu_liekf[:,:,0] = np.diag([(np.pi/180)**2, 0, 0])
    

# Creating initial theta given the heading_error

Xu_ekf[0,0] = - heading_error * (np.pi/180)
Xu_liekf[0,0] = - heading_error * (np.pi/180)

Xu_ekf[0,0] = -heading_error * (np.pi/180)
Xu_liekf[0,0] = -heading_error * (np.pi/180)

#-------------------------------------------------------------------

plt.figure(1)
plt.figure(1).clear()
plt.plot(X[1], X[2],"g", label = "True trajectory")

#Iterate through the filter list and calculate and plot trajectory for each filter

for config in filters:

    for k in range(1, n_time):

        Y = X[1:3,k]

        if config == "ekf":
            Xp, Pp = propagate(h, Xu_ekf[:,k-1].copy(), Pu_ekf[:,:,k-1].copy(), v, om, Q, config)
            Xu_ekf[:,k], Pu_ekf[:,:,k] = update(Xp, Pp, Y, N, config, k)
        elif config == "liekf":
            Xp, Pp = propagate(h, Xu_liekf[:,k-1].copy(), Pu_liekf[:,:,k-1].copy(), v, om, Q, config)
            Xu_liekf[:,k], Pu_liekf[:,:,k] = update(Xp, Pp, Y, N, config, k)
            
    
    if config == "ekf":
        plt.plot(Xu_ekf[1,:], Xu_ekf[2,:], "b--", label = "EKF - Estimated trajectory")
    
    elif config == "liekf":
        plt.plot(Xu_liekf[1,:], Xu_liekf[2,:], "r--", label = "LIEKF - Estimated trajectory")

plt.xlabel("x [m]")
plt.ylabel("y [m]")
plt.xlim([-10, 10])
plt.ylim([-2, 14])

plt.title(f"Attitude filtering, initial heading error = {heading_error}\N{DEGREE SIGN}")

plt.legend() 

plt.savefig(f"attitude_filtering_{heading_error}.png")
plt.show()

#-------------------------------------------------------------------

#Define the error of the estimated trajectories 

err_ekf = abs(X - Xu_ekf)
err_liekf = abs(X - Xu_liekf)

t = h*np.arange(n_time)

pos_err_liekf = np.zeros(n_time)
pos_err_ekf = np.zeros(n_time)
for i in range(n_time):
    pos_err_liekf[i] = err_liekf[1:3,i].sum()
    pos_err_ekf[i] = err_ekf[1:3,i].sum()
    
#-------------------------------------------------------------------

#Plot the attitude error

plt.figure(2)
plt.figure(2).clear()

plt.plot(t, err_liekf[0,:] / (np.pi/180), "r--", label = "LIEKF attitude error")
plt.plot(t, err_ekf[0,:] / (np.pi/180), "b--", label = "EKF attitude error")

plt.xlabel("time [s]")
plt.ylabel("Attitude error [degrees]")

plt.title(f"Attitude error, initial heading error = {heading_error}\N{DEGREE SIGN}")


plt.legend() 

plt.xlim([0, n_time*h])
plt.savefig(f"attitude_filtering_attitude_error_{heading_error}.png")
plt.show()

#-------------------------------------------------------------------

#Plot the position error

plt.figure(3)
plt.figure(3).clear()
    
plt.plot(t, pos_err_liekf, "r--", label = "LIEKF position error")
plt.plot(t, pos_err_ekf, "b--", label = "EKF position error")

plt.xlabel("time [s]")
plt.ylabel("Position error [m]")
plt.title(f"Position error, initial heading error = {heading_error}\N{DEGREE SIGN}")

plt.legend() 

plt.xlim([0, n_time*h])
plt.savefig(f"attitude_filtering_position_error_{heading_error}.png")
plt.show()